/*
 * [DomainStatus.java]
 *
 * Summary: Tracks whether a domain is valid and whether it has valid working mailservers. Keeps previous probes.
 *
 * Copyright: (c) 2002-2017 Roedy Green, Canadian Mind Products, http://mindprod.com
 *
 * Licence: This software may be copied and used freely for any purpose but military.
 *          http://mindprod.com/contact/nonmil.html
 *
 * Requires: JDK 1.8+
 *
 * Created with: JetBrains IntelliJ IDEA IDE http://www.jetbrains.com/idea/
 *
 * Version History:
 *  1.1 2002-08-07 new constants DEEP_FUTURE and DEEP_PAST to avoid wrap-around.
 */
package com.mindprod.bulk;

import java.io.Serializable;

import static com.mindprod.bulk.CustConfig.HISTORY_MASK;

/**
 * Tracks whether a domain is valid and whether it has valid working mailservers. Keeps previous probes.
 *
 * @author Roedy Green, Canadian Mind Products
 * @version 1.1 2002-08-07 new constants DEEP_FUTURE and DEEP_PAST to avoid wrap-around.
 * @since 2002-04-13
 */
public final class DomainStatus implements Serializable
    {
    /**
     * to prevent spurious InvalidClassExceptions.
     */
    static final long serialVersionUID = 1L;

    /**
     * A time in the far future. We don't use Long.MAX_VALUE because we may compare it with an interval around that
     * time, and we want to avoid wraparound.
     */
    private static final long DEEP_FUTURE =
            Long.MAX_VALUE - 100 * ( 365 * 24 * 60 * 60 * 1000L );

    /**
     * A time in the far past. We don't use Long.MIN_VALUE because we may compare it with an interval around that time,
     * and we want to avoid wraparound.
     */
    private static final long DEEP_PAST =
            Long.MIN_VALUE + 100/* years */ * ( 365 * 24 * 60 * 60 * 1000L );

    /**
     * name of domain, lower case whose status we are tracking.  e.g. mindprod.com or oberon.ark.com, not the mailserver
     * name, not an email address.
     */
    private final String domain;

    /**
     * on the most recent probe, did the mailserver exist in DNS records.  It need not have responded.
     */
    private boolean mailServerExists;

    /**
     * bit mask of previous test results from previous times we probed the mailserver being up. 1 = pass, 0 = fail. lsb
     * is oldest test. Tracks both DNS record checks and mail server probes.
     */
    private long history;

    /**
     * timestamp of last time we probed this domain to see how it's mailservers were doing. milliseconds since 1970 Jan
     * 1
     */
    private long lastProbe;

    /**
     * Last time inquired about his domain. Deadwood may be culled.
     */
    private long lastUsed;

    /**
     * Normal constructor
     *
     * @param domain name of domain, lower case whose status we will be tracking.  e.g. mindprod.com or oberon.ark.com,
     *               not the mailserver name, not an email address.
     */
    public DomainStatus( String domain )
        {
        this.domain = domain.trim().toLowerCase();
        this.history = 0;
        /* force reprobe soon */
        this.lastProbe = DEEP_PAST;
        this.lastUsed = DEEP_PAST;
        this.mailServerExists = false;
        }

    /**
     * constructor to create special exception domains.
     *
     * @param domain   name of domain, lower case whose status we will be tracking.  e.g. mindprod.com or
     *                 oberon.ark.com, not the mailserver name, not an email address.
     * @param howForce true means create DomainStatus that is always considered good no matter what the DNS records, or
     *                 results of serever probing say. false means create a DomainStatus that is always considered bad,
     *                 no matter what the evidence.
     */
    public DomainStatus( String domain, boolean howForce )
        {
        this.domain = domain.trim().toLowerCase();
        if ( howForce )
            {
            /* force to be considered always good */
            this.history = 0xffffffffffffffffL;
            this.lastProbe = DEEP_FUTURE;
            this.lastUsed = DEEP_FUTURE;
            this.mailServerExists = true;
            }
        else
            {
            /* force to be considered always bad */
            this.history = 0;
            this.lastProbe = DEEP_FUTURE;
            this.lastUsed = DEEP_FUTURE;
            this.mailServerExists = false;
            }
        }

    /**
     * Is it necessary to reprobe this domain to find its status.
     *
     * @return true if it is necessary to reprode. false if we can guess from past history.
     */
    private boolean isProbeNecessary()
        {
        long now = System.currentTimeMillis();
        if ( history < 0 )
            {
            /* i.e. high bit on, most recent test was pass */
            return lastProbe + CustConfig.PASSED_PROBE_INTERVAL < now;
            }
        else
            {
            /* i.e. high bit off, most recent test was fail */
            return lastProbe + CustConfig.FAILED_PROBE_INTERVAL < now;
            }
        }

    /**
     * Updates mailserverExists and history by examinging DNS records and probing the mailservers. Does probe whether
     * one is needed or not.
     */
    @SuppressWarnings( { "NestedAssignment" } )
    private void probe()
        {
        this.history >>>=
                1;/* slide over one reading, inserting default fail in MSB */
        this.lastProbe = System.currentTimeMillis();
        String[] mailServers = MailProber.getMailServersForDomain( domain );
        if ( ( this.mailServerExists = ( mailServers != null ) ) )
            {
            // list of mailservers from MX and A records.  Try them all for this domain.
            for ( String mailServer : mailServers )
                {
                if ( MailProber.isMailServerUp( mailServer ) )
                    {
                    this.history |=
                            0x8000000000000000L;/* turn on high bit, record good probe result */
                    return;
                    }
                }
            }
        /* no luck, try the domain itself as a mailserver */
        if ( MailProber.isMailServerUp( domain ) )
            {
            this.history |=
                    0x8000000000000000L;/* turn on high bit, record good probe result */
            }
        /* if fall out bottom, leave high order 0 bit= fail in probe history */
        } // end probe

    /**
     * Should this domain be culled from the cache?
     *
     * @return true if yes
     */
    boolean shouldCull()
        {
        long now = System.currentTimeMillis();
        return lastUsed + CustConfig.CULL_INTERVAL < now;
        }

    /**
     * Historically, does this domain exist, have mailservers and is at least one of those mailservers up and flying?
     * Examines past history, reprobes only as necessary. Good if: mailserver exists as of most recent probe. mailserver
     * responded recently to any probe.
     *
     * @return true if a mailserver is good.
     */
    public boolean isGood()
        {
        /* give a poke to keep this alive in cache, lastUSed, not lastProbe */
        this.lastUsed = System.currentTimeMillis();
        if ( isProbeNecessary() )
            {
            probe();
            }
        return history < 0 || mailServerExists && ( ( history & HISTORY_MASK )
                                                    != 0 );
        }
    }
