/*
 * [Log.java]
 *
 * Summary: Logs both to a file and to the console.
 *
 * Copyright: (c) 2007-2017 Roedy Green, Canadian Mind Products, http://mindprod.com
 *
 * Licence: This software may be copied and used freely for any purpose but military.
 *          http://mindprod.com/contact/nonmil.html
 *
 * Requires: JDK 1.8+
 *
 * Created with: JetBrains IntelliJ IDEA IDE http://www.jetbrains.com/idea/
 *
 * Version History:
 *  1.0 2007-05-27
 */
package com.mindprod.bulk;

import com.mindprod.common18.EIO;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.Reader;

import static java.lang.System.*;

/**
 * Logs both to a file and to the console.
 *
 * @author Roedy Green, Canadian Mind Products
 * @version 1.0 2007-05-27
 * @since 2007-05-27
 */
final class Log
    {
    /**
     * Same of the log file .
     */
    private final String logFilename;

    /**
     * Where the log entries are written.
     */
    private PrintWriter logFile;

    /**
     * constructor.
     *
     * @param logFilename name of file where to write log entries.
     */
    public Log( String logFilename )
        {
        this.logFilename = logFilename;
        }

    /**
     * Get log as String. Also clears log.
     *
     * @return entire log as a single string with embedded CrLfs.
     * @throws IOException if cannot read the log.
     */
    public String asString() throws IOException
        {
        close();
        final File f = new File( logFilename );
        Reader r = new FileReader( logFilename );
        char[] ca = new char[ ( int ) f.length() ];
        int charsRead = r.read( ca );
        String s = new String( ca, 0, charsRead );
        open();
        return s;
        }

    /**
     * close the log.
     */
    public void close()
        {
        if ( logFile != null )
            {
            logFile.close();
            logFile = null;
            }
        }

    /**
     * Open the log if not already open.
     *
     * @throws IOException if can't open log.
     */
    public void open() throws IOException
        {
        if ( logFile == null )
            {
            logFile = EIO.getPrintWriter( new File( logFilename ), 4 * 1024, EIO.UTF8 );
            }
        }

    /**
     * Log the string both to file and to console.
     *
     * @param s string to log.
     */
    public void println( String s )
        {
        out.println( s );
        logFile.println( s );
        }

    /**
     * Log the string.
     *
     * @param s             string to log
     * @param echoToConsole true if string should be echoed to console as well.
     */
    @SuppressWarnings( { "UnusedDeclaration" } )
    public void println( String s, boolean echoToConsole )
        {
        if ( echoToConsole )
            {
            out.println( s );
            }
        logFile.println( s );
        }
    }
