/*
 * [MailProber.java]
 *
 * Summary: probe a mailserver to see if it is valid and functioning.
 *
 * Copyright: (c) 2002-2017 Roedy Green, Canadian Mind Products, http://mindprod.com
 *
 * Licence: This software may be copied and used freely for any purpose but military.
 *          http://mindprod.com/contact/nonmil.html
 *
 * Requires: JDK 1.8+
 *
 * Created with: JetBrains IntelliJ IDEA IDE http://www.jetbrains.com/idea/
 *
 * Version History:
 *  1.7 2007-08-21
 */
package com.mindprod.bulk;

import javax.naming.NamingException;
import javax.naming.directory.Attribute;
import javax.naming.directory.Attributes;
import javax.naming.directory.DirContext;
import javax.naming.directory.InitialDirContext;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.net.Socket;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.Hashtable;

import static com.mindprod.bulk.CustConfig.DEBUGGING;
import static com.mindprod.bulk.CustConfig.PROBE_DOMAIN;
import static com.mindprod.bulk.CustConfig.PROBE_TIMEOUT;
import static java.lang.System.*;

/**
 * probe a mailserver to see if it is valid and functioning.
 *
 * @author Roedy Green, Canadian Mind Products
 * @version 1.7 2007-08-21
 * @since 2002
 */
public final class MailProber
    {
    private static final int STANDARD_SMTP_PORT = 25;

    private static final MXComparator mxComparator = new MXComparator();

    /**
     * Gets all matching dns records as an array of strings.
     *
     * @param domain domain, e.g. oberon.ark.com or oberon.com which you want the DNS records.
     * @param types  e.g."MX","A" to describe which types of record you want.
     *
     * @return ArrayList of Strings
     * @throws NamingException if DNS lookup fails.
     */
    @SuppressWarnings( { "PointlessBooleanExpression", "ConstantConditions" } )
    private static ArrayList<String> getDNSRecs( String domain,
                                                 String... types ) throws NamingException
        {
        ArrayList<String> results = new ArrayList<>( 15 );
        //        Old style required you to provide an explicit DNS server.
        //        DirContext ictx = new InitialDirContext();
        //        Attributes attrs =
        //                ictx.getAttributes( "dns://" + DNS_SERVER + "/" + domain,
        //                                    types );
        // Don't change to HashMap. Legacy interface.
        Hashtable<String, String> env = new Hashtable<>( 11 );
        env.put( "java.naming.factory.initial",
                "com.sun.jndi.dns.DnsContextFactory" );
        DirContext ictx = new InitialDirContext( env );
        Attributes attrs = ictx.getAttributes( domain, types );
        for ( Enumeration e = attrs.getAll(); e.hasMoreElements(); )
            {
            Attribute a = ( Attribute ) e.nextElement();
            int size = a.size();
            for ( int i = 0; i < size; i++ )
                {
                // MX string has priority (lower better) followed by associated mailserver
                // A string is just IP
                results.add( ( String ) a.get( i ) );
                } // end inner for
            } // end  outer for
        if ( DEBUGGING && results.size() == 0 )
            {
            err.println( "Failed to find any DNS records for domain "
                         + domain );
            }
        return results;
        }

    /**
     * Get all mailservers for a given domain.
     *
     * @param domain name of domain, e.g. mindprod.com or oberon.ark.com
     *
     * @return array of strings of names of mailservers, may be DNS names or ip names, suitable for
     * InternetAddress.getByName. Null if none.
     */
    @SuppressWarnings( { "ValueOfIncrementOrDecrementUsed" } )
    public static String[] getMailServersForDomain( String domain )
        {
        try
            {
            domain = domain.trim().toLowerCase();
            int j = 0;
            ArrayList<String> priosAndNames = getDNSRecs( domain, "MX" );
            // should look like 99 domain.com with trailing .
            // "5 gmail-smtp-in.l.google.com."
            // "10 alt1.gmail-smtp-in.l.google.com."
            // sometimes looks like domain.com.  without leading priority.
            Collections.sort( priosAndNames,
                    mxComparator );/* prio 0 the best comes first */
            ArrayList<String> ips = getDNSRecs( domain, "A" );
            // should just be an IP per line.
            if ( CustConfig.DEBUGGING )
                {
                out.println( "MX records for domain " + domain );
                for ( String pn : priosAndNames )
                    {
                    out.println( "\"" + pn + "\"" );
                    }
                out.println( "A records" );
                for ( String ip : ips )
                    {
                    out.println( ip );
                    }
                }
            String[] results = new String[ priosAndNames.size() + ips.size() ];
            /* copy MX records, chopping prio */
            for ( String prioAndName : priosAndNames )
                {
                int spacePlace = prioAndName.indexOf( ' ' );
                // if space not found, get whole string.
                prioAndName = prioAndName.substring( spacePlace + 1 );
                results[ j++ ] = prioAndName;
                }
            /* copy over A records as is */
            for ( String ip : ips )
                {
                results[ j++ ] = ip;
                }
            if ( j == 0 )
                {
                return null;
                }
            return results;
            } // end try
        catch ( NamingException e )
            {
            return null;
            }
        }

    /**
     * Probe mailserver to see if it is alive and accepting SMTP mail.
     *
     * @param mailserver Fully qualified name of mailserver, e.g. mail.mindprod.com may also be an IP.
     *
     * @return true if mailserver at least responds.
     */
    public static boolean isMailServerUp( String mailserver )
        {
        try
            {
            // O P E N
            // Mailserver may accept SMTP connect on alternate sockets, but always
            // accepts on port 25.
            final Socket socket = new Socket( mailserver, STANDARD_SMTP_PORT
                    /* always socket # 25*/ );
            socket.setSoTimeout( PROBE_TIMEOUT );/* does not appear to work on W2K */
            final OutputStream os = socket.getOutputStream();
            OutputStreamWriter osw = new OutputStreamWriter( os );
            PrintWriter prw = new PrintWriter( osw, true/* auto flush */ );
            final InputStream is = socket.getInputStream();
            prw.print( "HELO " + PROBE_DOMAIN + "\r\n" );
            prw.flush();
            // R E A D
            // possible valid responses
            // 220 = initial greeting
            // 250 = ok to my HELO
            byte[] response = new byte[ 3 ];
            int bytesRead = is.read( response, 0/* offset */, response.length );
            // say goodbye so they will recognise our next HELO
            prw.print( "QUIT\r\n" );
            prw.flush();
            // C L O S E
            prw.close();
            is.close();
            if ( bytesRead < 3 )
                {
                return false;
                }
            /* get first three chars */
            String r = new String( response, 0, 3 );
            if ( DEBUGGING )
                {
                out.println( "probe " + mailserver + " response: " + r );
                }
            return r.equals( "220" ) || r.equals( "250" );
            }
        catch ( Exception e )
            {
            /* timeout most likely */
            return false;
            }
        }

    /**
     * Test driver
     *
     * @param args list of domains to check.
     */
    public static void main( String[] args )
        {
        for ( String domain : args )
            {
            out.println( "domain: " + domain );
            String[] s = getMailServersForDomain( domain );
            if ( s != null )
                {
                for ( String value : s )
                    {
                    out.println( value );
                    }
                }
            }
        }
    }
